@@ -1,7 +1,7 @@
-/*	$Id: eqn.c,v 1.78 2017/07/15 16:26:17 schwarze Exp $ */
+/*	$Id: eqn.c,v 1.83 2018/12/14 06:33:14 schwarze Exp $ */
 /*
  * Copyright (c) 2011, 2014 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2014, 2015, 2017 Ingo Schwarze <schwarze@openbsd.org>
+ * Copyright (c) 2014, 2015, 2017, 2018 Ingo Schwarze <schwarze@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -30,8 +30,9 @@
 #include "mandoc_aux.h"
 #include "mandoc.h"
 #include "roff.h"
+#include "eqn.h"
 #include "libmandoc.h"
-#include "libroff.h"
+#include "eqn_parse.h"
 
 #define	EQN_NEST_MAX	 128 /* maximum nesting of defines */
 #define	STRNEQ(p1, sz1, p2, sz2) \
@@ -284,6 +285,13 @@ enum	parse_mode {
 	MODE_TOK
 };
 
+struct	eqn_def {
+	char		 *key;
+	size_t		  keysz;
+	char		 *val;
+	size_t		  valsz;
+};
+
 static	struct eqn_box	*eqn_box_alloc(struct eqn_node *, struct eqn_box *);
 static	struct eqn_box	*eqn_box_makebinary(struct eqn_node *,
 				struct eqn_box *);
@@ -295,12 +303,11 @@ static	void		 eqn_undef(struct eqn_node *);
 
 
 struct eqn_node *
-eqn_alloc(struct mparse *parse)
+eqn_alloc(void)
 {
 	struct eqn_node *ep;
 
 	ep = mandoc_calloc(1, sizeof(*ep));
-	ep->parse = parse;
 	ep->gsize = EQN_DEFSIZE;
 	return ep;
 }
@@ -399,7 +406,7 @@ eqn_next(struct eqn_node *ep, enum parse_mode mode)
 			ep->end = strchr(ep->start + 1, *ep->start);
 			ep->start++;  /* Skip opening quote. */
 			if (ep->end == NULL) {
-				mandoc_msg(MANDOCERR_ARG_QUOTE, ep->parse,
+				mandoc_msg(MANDOCERR_ARG_QUOTE,
 				    ep->node->line, ep->node->pos, NULL);
 				ep->end = strchr(ep->start, '\0');
 			}
@@ -420,7 +427,7 @@ eqn_next(struct eqn_node *ep, enum parse_mode mode)
 		if ((def = eqn_def_find(ep)) == NULL)
 			break;
 		if (++lim > EQN_NEST_MAX) {
-			mandoc_msg(MANDOCERR_ROFFLOOP, ep->parse,
+			mandoc_msg(MANDOCERR_ROFFLOOP,
 			    ep->node->line, ep->node->pos, NULL);
 			return EQN_TOK_EOF;
 		}
@@ -468,6 +475,8 @@ eqn_next(struct eqn_node *ep, enum parse_mode mode)
 void
 eqn_box_free(struct eqn_box *bp)
 {
+	if (bp == NULL)
+		return;
 
 	if (bp->first)
 		eqn_box_free(bp->first);
@@ -482,6 +491,16 @@ eqn_box_free(struct eqn_box *bp)
 	free(bp);
 }
 
+struct eqn_box *
+eqn_box_new(void)
+{
+	struct eqn_box	*bp;
+
+	bp = mandoc_calloc(1, sizeof(*bp));
+	bp->expectargs = UINT_MAX;
+	return bp;
+}
+
 /*
  * Allocate a box as the last child of the parent node.
  */
@@ -490,10 +509,9 @@ eqn_box_alloc(struct eqn_node *ep, struct eqn_box *parent)
 {
 	struct eqn_box	*bp;
 
-	bp = mandoc_calloc(1, sizeof(struct eqn_box));
+	bp = eqn_box_new();
 	bp->parent = parent;
 	bp->parent->args++;
-	bp->expectargs = UINT_MAX;
 	bp->font = bp->parent->font;
 	bp->size = ep->gsize;
 
@@ -542,7 +560,7 @@ static void
 eqn_delim(struct eqn_node *ep)
 {
 	if (ep->end[0] == '\0' || ep->end[1] == '\0') {
-		mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
+		mandoc_msg(MANDOCERR_REQ_EMPTY,
 		    ep->node->line, ep->node->pos, "delim");
 		if (ep->end[0] != '\0')
 			ep->end++;
@@ -569,7 +587,7 @@ eqn_undef(struct eqn_node *ep)
 	struct eqn_def	*def;
 
 	if (eqn_next(ep, MODE_NOSUB) == EQN_TOK_EOF) {
-		mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
+		mandoc_msg(MANDOCERR_REQ_EMPTY,
 		    ep->node->line, ep->node->pos, "undef");
 		return;
 	}
@@ -588,7 +606,7 @@ eqn_def(struct eqn_node *ep)
 	int		 i;
 
 	if (eqn_next(ep, MODE_NOSUB) == EQN_TOK_EOF) {
-		mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
+		mandoc_msg(MANDOCERR_REQ_EMPTY,
 		    ep->node->line, ep->node->pos, "define");
 		return;
 	}
@@ -617,7 +635,7 @@ eqn_def(struct eqn_node *ep)
 	}
 
 	if (eqn_next(ep, MODE_QUOTED) == EQN_TOK_EOF) {
-		mandoc_vmsg(MANDOCERR_REQ_EMPTY, ep->parse,
+		mandoc_msg(MANDOCERR_REQ_EMPTY,
 		    ep->node->line, ep->node->pos, "define %s", def->key);
 		free(def->key);
 		free(def->val);
@@ -666,7 +684,7 @@ next_tok:
 	case EQN_TOK_TDEFINE:
 		if (eqn_next(ep, MODE_NOSUB) == EQN_TOK_EOF ||
 		    eqn_next(ep, MODE_QUOTED) == EQN_TOK_EOF)
-			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
+			mandoc_msg(MANDOCERR_REQ_EMPTY,
 			    ep->node->line, ep->node->pos, "tdefine");
 		break;
 	case EQN_TOK_DELIM:
@@ -674,8 +692,8 @@ next_tok:
 		break;
 	case EQN_TOK_GFONT:
 		if (eqn_next(ep, MODE_SUB) == EQN_TOK_EOF)
-			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 		break;
 	case EQN_TOK_MARK:
 	case EQN_TOK_LINEUP:
@@ -690,8 +708,8 @@ next_tok:
 	case EQN_TOK_DOT:
 	case EQN_TOK_DOTDOT:
 		if (parent->last == NULL) {
-			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			cur = eqn_box_alloc(ep, parent);
 			cur->type = EQN_TEXT;
 			cur->text = mandoc_strdup("");
@@ -735,8 +753,8 @@ next_tok:
 	case EQN_TOK_DOWN:
 	case EQN_TOK_UP:
 		if (eqn_next(ep, MODE_SUB) == EQN_TOK_EOF)
-			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 		break;
 	case EQN_TOK_FAT:
 	case EQN_TOK_ROMAN:
@@ -773,14 +791,14 @@ next_tok:
 	case EQN_TOK_GSIZE:
 		/* Accept two values: integral size and a single. */
 		if (eqn_next(ep, MODE_SUB) == EQN_TOK_EOF) {
-			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			break;
 		}
 		size = mandoc_strntoi(ep->start, ep->toksz, 10);
 		if (-1 == size) {
-			mandoc_msg(MANDOCERR_IT_NONUM, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_IT_NONUM, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			break;
 		}
 		if (EQN_TOK_GSIZE == tok) {
@@ -804,8 +822,8 @@ next_tok:
 		 * and keep on reading.
 		 */
 		if (parent->last == NULL) {
-			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			cur = eqn_box_alloc(ep, parent);
 			cur->type = EQN_TEXT;
 			cur->text = mandoc_strdup("");
@@ -871,8 +889,8 @@ next_tok:
 		 * rebalance and continue reading.
 		 */
 		if (parent->last == NULL) {
-			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_EQN_NOBOX, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			cur = eqn_box_alloc(ep, parent);
 			cur->type = EQN_TEXT;
 			cur->text = mandoc_strdup("");
@@ -898,16 +916,16 @@ next_tok:
 			     cur->left != NULL))
 				break;
 		if (cur == NULL) {
-			mandoc_msg(MANDOCERR_BLK_NOTOPEN, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_BLK_NOTOPEN, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			break;
 		}
 		parent = cur;
 		if (EQN_TOK_RIGHT == tok) {
 			if (eqn_next(ep, MODE_SUB) == EQN_TOK_EOF) {
 				mandoc_msg(MANDOCERR_REQ_EMPTY,
-				    ep->parse, ep->node->line,
-				    ep->node->pos, eqn_toks[tok]);
+				    ep->node->line, ep->node->pos,
+				    "%s", eqn_toks[tok]);
 				break;
 			}
 			/* Handling depends on right/left. */
@@ -941,8 +959,8 @@ next_tok:
 			parent = parent->parent;
 		if (EQN_TOK_LEFT == tok &&
 		    eqn_next(ep, MODE_SUB) == EQN_TOK_EOF) {
-			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_REQ_EMPTY, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			break;
 		}
 		parent = eqn_box_alloc(ep, parent);
@@ -975,8 +993,8 @@ next_tok:
 			if (cur->type == EQN_PILE)
 				break;
 		if (cur == NULL) {
-			mandoc_msg(MANDOCERR_IT_STRAY, ep->parse,
-			    ep->node->line, ep->node->pos, eqn_toks[tok]);
+			mandoc_msg(MANDOCERR_IT_STRAY, ep->node->line,
+			    ep->node->pos, "%s", eqn_toks[tok]);
 			break;
 		}
 		parent = eqn_box_alloc(ep, cur);
@@ -1092,6 +1110,9 @@ eqn_free(struct eqn_node *p)
 {
 	int		 i;
 
+	if (p == NULL)
+		return;
+
 	for (i = 0; i < (int)p->defsz; i++) {
 		free(p->defs[i].key);
 		free(p->defs[i].val);
