@@ -1,4 +1,4 @@
-/*	$Id: eqn_term.c,v 1.17 2017/08/23 21:56:20 schwarze Exp $ */
+/*	$Id: eqn_term.c,v 1.19 2018/12/13 05:23:38 schwarze Exp $ */
 /*
  * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
  * Copyright (c) 2014, 2015, 2017 Ingo Schwarze <schwarze@openbsd.org>
@@ -25,7 +25,7 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include "mandoc.h"
+#include "eqn.h"
 #include "out.h"
 #include "term.h"
 
@@ -106,7 +106,7 @@ eqn_box(struct termp *p, const struct eqn_box *bp)
 	/* Special box types. */
 
 	if (bp->pos == EQNPOS_SQRT) {
-		term_word(p, "sqrt");
+		term_word(p, "\\(sr");
 		if (bp->first != NULL) {
 			p->flags |= TERMP_NOSPACE;
 			eqn_box(p, bp->first);
