@@ -1,7 +1,7 @@
-/*	$Id: main.c,v 1.306 2018/05/14 14:10:23 schwarze Exp $ */
+/*	$Id: main.c,v 1.322 2019/03/06 10:18:58 schwarze Exp $ */
 /*
  * Copyright (c) 2008-2012 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2010-2012, 2014-2018 Ingo Schwarze <schwarze@openbsd.org>
+ * Copyright (c) 2010-2012, 2014-2019 Ingo Schwarze <schwarze@openbsd.org>
  * Copyright (c) 2010 Joerg Sonnenberger <joerg@netbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -21,7 +21,6 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/param.h>	/* MACHINE */
-#include <sys/termios.h>
 #include <sys/wait.h>
 
 #include <assert.h>
@@ -40,6 +39,7 @@
 #include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+#include <termios.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -49,6 +49,7 @@
 #include "roff.h"
 #include "mdoc.h"
 #include "man.h"
+#include "mandoc_parse.h"
 #include "tag.h"
 #include "main.h"
 #include "manconf.h"
@@ -81,7 +82,6 @@ struct	curparse {
 	void		 *outdata;	/* data for output */
 	char		 *os_s;		/* operating system for display */
 	int		  wstop;	/* stop after a file with a warning */
-	enum mandocerr	  mmin;		/* ignore messages below this */
 	enum mandoc_os	  os_e;		/* check base system conventions */
 	enum outt	  outtype;	/* which output to use */
 };
@@ -89,7 +89,7 @@ struct	curparse {
 
 int			  mandocdb(int, char *[]);
 
-static	void		  check_xr(const char *);
+static	void		  check_xr(void);
 static	int		  fs_lookup(const struct manpaths *,
 				size_t ipath, const char *,
 				const char *, const char *,
@@ -99,8 +99,6 @@ static	int		  fs_search(const struct mansearch *,
 				struct manpage **, size_t *);
 static	int		  koptions(int *, char *);
 static	void		  moptions(int *, char *);
-static	void		  mmsg(enum mandocerr, enum mandoclevel,
-				const char *, int, int, const char *);
 static	void		  outdata_alloc(struct curparse *);
 static	void		  parse(struct curparse *, int, const char *);
 static	void		  passthrough(const char *, int, int);
@@ -112,8 +110,6 @@ static	int		  woptions(struct curparse *, char *);
 static	const int sec_prios[] = {1, 4, 5, 8, 6, 3, 7, 2, 9};
 static	char		  help_arg[] = "help";
 static	char		 *help_argv[] = {help_arg, NULL};
-static	enum mandoclevel  rc;
-static	FILE		 *mmsg_stream;
 
 
 int
@@ -127,7 +123,7 @@ main(int argc, char *argv[])
 	struct manpage	*res, *resp;
 	const char	*progname, *sec, *thisarg;
 	char		*conf_file, *defpaths, *auxpaths;
-	char		*oarg;
+	char		*oarg, *tagarg;
 	unsigned char	*uc;
 	size_t		 i, sz;
 	int		 prio, best_prio;
@@ -152,6 +148,7 @@ main(int argc, char *argv[])
 	setprogname(progname);
 #endif
 
+	mandoc_msg_setoutfile(stderr);
 	if (strncmp(progname, "mandocdb", 8) == 0 ||
 	    strcmp(progname, BINM_MAKEWHATIS) == 0)
 		return mandocdb(argc, argv);
@@ -191,10 +188,8 @@ main(int argc, char *argv[])
 
 	memset(&curp, 0, sizeof(struct curparse));
 	curp.outtype = OUTT_LOCALE;
-	curp.mmin = MANDOCERR_MAX;
 	curp.outopts = &conf.output;
 	options = MPARSE_SO | MPARSE_UTF8 | MPARSE_LATIN1;
-	mmsg_stream = stderr;
 
 	use_pager = 1;
 	tag_files = NULL;
@@ -314,6 +309,9 @@ main(int argc, char *argv[])
 		}
 	}
 
+	if (curp.outtype != OUTT_TREE || !curp.outopts->noval)
+		options |= MPARSE_VALIDATE;
+
 	if (outmode == OUTMODE_FLN ||
 	    outmode == OUTMODE_LST ||
 	    !isatty(STDOUT_FILENO))
@@ -371,7 +369,16 @@ main(int argc, char *argv[])
 #endif
 	}
 
-	rc = MANDOCLEVEL_OK;
+	/*
+	 * Use the first argument for -O tag in addition to
+	 * using it as a search term for man(1) or apropos(1).
+	 */
+
+	if (conf.output.tag != NULL && *conf.output.tag == '\0') {
+		tagarg = argc > 0 && search.argmode == ARG_EXPR ?
+		    strchr(*argv, '=') : NULL;
+		conf.output.tag = tagarg == NULL ? *argv : tagarg + 1;
+	}
 
 	/* man(1), whatis(1), apropos(1) */
 
@@ -405,7 +412,6 @@ main(int argc, char *argv[])
 				res[sz].names = NULL;
 				res[sz].output = NULL;
 				res[sz].ipath = SIZE_MAX;
-				res[sz].bits = 0;
 				res[sz].sec = 10;
 				res[sz].form = FORM_SRC;
 				sz++;
@@ -415,7 +421,7 @@ main(int argc, char *argv[])
 		if (sz == 0) {
 			if (search.argmode != ARG_NAME)
 				warnx("nothing appropriate");
-			rc = MANDOCLEVEL_BADARG;
+			mandoc_msg_setrc(MANDOCLEVEL_BADARG);
 			goto out;
 		}
 
@@ -483,19 +489,17 @@ main(int argc, char *argv[])
 		moptions(&options, auxpaths);
 
 	mchars_alloc();
-	curp.mp = mparse_alloc(options, curp.mmin, mmsg,
-	    curp.os_e, curp.os_s);
-
-	/*
-	 * Conditionally start up the lookaside buffer before parsing.
-	 */
-	if (OUTT_MAN == curp.outtype)
-		mparse_keep(curp.mp);
+	curp.mp = mparse_alloc(options, curp.os_e, curp.os_s);
 
 	if (argc < 1) {
-		if (use_pager)
+		if (use_pager) {
 			tag_files = tag_init();
-		parse(&curp, STDIN_FILENO, "<stdin>");
+			tag_files->tagname = conf.output.tag;
+		}
+		thisarg = "<stdin>";
+		mandoc_msg_setinfilename(thisarg);
+		parse(&curp, STDIN_FILENO, thisarg);
+		mandoc_msg_setinfilename(NULL);
 	}
 
 	/*
@@ -519,22 +523,25 @@ main(int argc, char *argv[])
 				(void)chdir(conf.manpath.paths[resp->ipath]);
 			else if (startdir != -1)
 				(void)fchdir(startdir);
-		}
+			thisarg = resp->file;
+		} else
+			thisarg = *argv;
 
-		fd = mparse_open(curp.mp, resp != NULL ? resp->file : *argv);
+		fd = mparse_open(curp.mp, thisarg);
 		if (fd != -1) {
 			if (use_pager) {
-				tag_files = tag_init();
 				use_pager = 0;
+				tag_files = tag_init();
+				tag_files->tagname = conf.output.tag;
 			}
 
-			if (resp == NULL)
-				parse(&curp, fd, *argv);
-			else if (resp->form == FORM_SRC)
-				parse(&curp, fd, resp->file);
+			mandoc_msg_setinfilename(thisarg);
+			if (resp == NULL || resp->form == FORM_SRC)
+				parse(&curp, fd, thisarg);
 			else
 				passthrough(resp->file, fd,
 				    conf.output.synopsisonly);
+			mandoc_msg_setinfilename(NULL);
 
 			if (ferror(stdout)) {
 				if (tag_files != NULL) {
@@ -543,7 +550,7 @@ main(int argc, char *argv[])
 					tag_files = NULL;
 				} else
 					warn("stdout");
-				rc = MANDOCLEVEL_SYSERR;
+				mandoc_msg_setrc(MANDOCLEVEL_SYSERR);
 				break;
 			}
 
@@ -552,10 +559,11 @@ main(int argc, char *argv[])
 					outdata_alloc(&curp);
 				terminal_sepline(curp.outdata);
 			}
-		} else if (rc < MANDOCLEVEL_ERROR)
-			rc = MANDOCLEVEL_ERROR;
+		} else
+			mandoc_msg(MANDOCERR_FILE, 0, 0,
+			    "%s: %s", thisarg, strerror(errno));
 
-		if (MANDOCLEVEL_OK != rc && curp.wstop)
+		if (curp.wstop && mandoc_msg_getrc() != MANDOCLEVEL_OK)
 			break;
 
 		if (resp != NULL)
@@ -646,7 +654,7 @@ out:
 
 			if (pid == -1) {
 				warn("wait");
-				rc = MANDOCLEVEL_SYSERR;
+				mandoc_msg_setrc(MANDOCLEVEL_SYSERR);
 				break;
 			}
 			if (!WIFSTOPPED(status))
@@ -656,8 +664,7 @@ out:
 		}
 		tag_unlink();
 	}
-
-	return (int)rc;
+	return (int)mandoc_msg_getrc();
 }
 
 static void
@@ -755,7 +762,6 @@ found:
 	page->names = NULL;
 	page->output = NULL;
 	page->ipath = ipath;
-	page->bits = NAME_FILE & NAME_MASK;
 	page->sec = (*sec >= '1' && *sec <= '9') ? *sec - '1' + 1 : 10;
 	page->form = form;
 	return 1;
@@ -790,8 +796,18 @@ fs_search(const struct mansearch *cfg, const struct manpaths *paths,
 					return 1;
 		}
 		if (res != NULL && *ressz == lastsz &&
-		    strchr(*argv, '/') == NULL)
-			warnx("No entry for %s in the manual.", *argv);
+		    strchr(*argv, '/') == NULL) {
+			if (cfg->arch != NULL &&
+			    arch_valid(cfg->arch, OSENUM) == 0)
+				warnx("Unknown architecture \"%s\".",
+				    cfg->arch);
+			else if (cfg->sec == NULL)
+				warnx("No entry for %s in the manual.",
+				    *argv);
+			else
+				warnx("No entry for %s in section %s "
+				    "of the manual.", *argv, cfg->sec);
+		}
 		lastsz = *ressz;
 		argv++;
 		argc--;
@@ -802,101 +818,92 @@ fs_search(const struct mansearch *cfg, const struct manpaths *paths,
 static void
 parse(struct curparse *curp, int fd, const char *file)
 {
-	enum mandoclevel  rctmp;
-	struct roff_man	 *man;
+	struct roff_meta *meta;
 
 	/* Begin by parsing the file itself. */
 
 	assert(file);
 	assert(fd >= 0);
 
-	rctmp = mparse_readfd(curp->mp, fd, file);
+	mparse_readfd(curp->mp, fd, file);
 	if (fd != STDIN_FILENO)
 		close(fd);
-	if (rc < rctmp)
-		rc = rctmp;
 
 	/*
 	 * With -Wstop and warnings or errors of at least the requested
 	 * level, do not produce output.
 	 */
 
-	if (rctmp != MANDOCLEVEL_OK && curp->wstop)
+	if (curp->wstop && mandoc_msg_getrc() != MANDOCLEVEL_OK)
 		return;
 
 	if (curp->outdata == NULL)
 		outdata_alloc(curp);
+	else if (curp->outtype == OUTT_HTML)
+		html_reset(curp);
 
-	mparse_result(curp->mp, &man, NULL);
+	mandoc_xr_reset();
+	meta = mparse_result(curp->mp);
 
 	/* Execute the out device, if it exists. */
 
-	if (man == NULL)
-		return;
-	mandoc_xr_reset();
-	if (man->macroset == MACROSET_MDOC) {
-		if (curp->outtype != OUTT_TREE || !curp->outopts->noval)
-			mdoc_validate(man);
+	if (meta->macroset == MACROSET_MDOC) {
 		switch (curp->outtype) {
 		case OUTT_HTML:
-			html_mdoc(curp->outdata, man);
+			html_mdoc(curp->outdata, meta);
 			break;
 		case OUTT_TREE:
-			tree_mdoc(curp->outdata, man);
+			tree_mdoc(curp->outdata, meta);
 			break;
 		case OUTT_MAN:
-			man_mdoc(curp->outdata, man);
+			man_mdoc(curp->outdata, meta);
 			break;
 		case OUTT_PDF:
 		case OUTT_ASCII:
 		case OUTT_UTF8:
 		case OUTT_LOCALE:
 		case OUTT_PS:
-			terminal_mdoc(curp->outdata, man);
+			terminal_mdoc(curp->outdata, meta);
 			break;
 		case OUTT_MARKDOWN:
-			markdown_mdoc(curp->outdata, man);
+			markdown_mdoc(curp->outdata, meta);
 			break;
 		default:
 			break;
 		}
 	}
-	if (man->macroset == MACROSET_MAN) {
-		if (curp->outtype != OUTT_TREE || !curp->outopts->noval)
-			man_validate(man);
+	if (meta->macroset == MACROSET_MAN) {
 		switch (curp->outtype) {
 		case OUTT_HTML:
-			html_man(curp->outdata, man);
+			html_man(curp->outdata, meta);
 			break;
 		case OUTT_TREE:
-			tree_man(curp->outdata, man);
+			tree_man(curp->outdata, meta);
 			break;
 		case OUTT_MAN:
-			man_man(curp->outdata, man);
+			mparse_copy(curp->mp);
 			break;
 		case OUTT_PDF:
 		case OUTT_ASCII:
 		case OUTT_UTF8:
 		case OUTT_LOCALE:
 		case OUTT_PS:
-			terminal_man(curp->outdata, man);
+			terminal_man(curp->outdata, meta);
 			break;
 		default:
 			break;
 		}
 	}
-	if (curp->mmin < MANDOCERR_STYLE)
-		check_xr(file);
-	mparse_updaterc(curp->mp, &rc);
+	if (mandoc_msg_getmin() < MANDOCERR_STYLE)
+		check_xr();
 }
 
 static void
-check_xr(const char *file)
+check_xr(void)
 {
 	static struct manpaths	 paths;
 	struct mansearch	 search;
 	struct mandoc_xr	*xr;
-	char			*cp;
 	size_t			 sz;
 
 	if (paths.sz == 0)
@@ -915,13 +922,12 @@ check_xr(const char *file)
 		if (fs_search(&search, &paths, 1, &xr->name, NULL, &sz))
 			continue;
 		if (xr->count == 1)
-			mandoc_asprintf(&cp, "Xr %s %s", xr->name, xr->sec);
+			mandoc_msg(MANDOCERR_XR_BAD, xr->line,
+			    xr->pos + 1, "Xr %s %s", xr->name, xr->sec);
 		else
-			mandoc_asprintf(&cp, "Xr %s %s (%d times)",
+			mandoc_msg(MANDOCERR_XR_BAD, xr->line,
+			    xr->pos + 1, "Xr %s %s (%d times)",
 			    xr->name, xr->sec, xr->count);
-		mmsg(MANDOCERR_XR_BAD, MANDOCLEVEL_STYLE,
-		    file, xr->line, xr->pos + 1, cp);
-		free(cp);
 	}
 }
 
@@ -1020,8 +1026,7 @@ done:
 fail:
 	free(line);
 	warn("%s: SYSERR: %s", file, syscall);
-	if (rc < MANDOCLEVEL_SYSERR)
-		rc = MANDOCLEVEL_SYSERR;
+	mandoc_msg_setrc(MANDOCLEVEL_SYSERR);
 }
 
 static int
@@ -1063,8 +1068,8 @@ toptions(struct curparse *curp, char *arg)
 		curp->outtype = OUTT_ASCII;
 	else if (0 == strcmp(arg, "lint")) {
 		curp->outtype = OUTT_LINT;
-		curp->mmin = MANDOCERR_BASE;
-		mmsg_stream = stdout;
+		mandoc_msg_setoutfile(stdout);
+		mandoc_msg_setmin(MANDOCERR_BASE);
 	} else if (0 == strcmp(arg, "tree"))
 		curp->outtype = OUTT_TREE;
 	else if (0 == strcmp(arg, "man"))
@@ -1115,29 +1120,29 @@ woptions(struct curparse *curp, char *arg)
 			break;
 		case 1:
 		case 2:
-			curp->mmin = MANDOCERR_BASE;
+			mandoc_msg_setmin(MANDOCERR_BASE);
 			break;
 		case 3:
-			curp->mmin = MANDOCERR_STYLE;
+			mandoc_msg_setmin(MANDOCERR_STYLE);
 			break;
 		case 4:
-			curp->mmin = MANDOCERR_WARNING;
+			mandoc_msg_setmin(MANDOCERR_WARNING);
 			break;
 		case 5:
-			curp->mmin = MANDOCERR_ERROR;
+			mandoc_msg_setmin(MANDOCERR_ERROR);
 			break;
 		case 6:
-			curp->mmin = MANDOCERR_UNSUPP;
+			mandoc_msg_setmin(MANDOCERR_UNSUPP);
 			break;
 		case 7:
-			curp->mmin = MANDOCERR_MAX;
+			mandoc_msg_setmin(MANDOCERR_MAX);
 			break;
 		case 8:
-			curp->mmin = MANDOCERR_BASE;
+			mandoc_msg_setmin(MANDOCERR_BASE);
 			curp->os_e = MANDOC_OS_OPENBSD;
 			break;
 		case 9:
-			curp->mmin = MANDOCERR_BASE;
+			mandoc_msg_setmin(MANDOCERR_BASE);
 			curp->os_e = MANDOC_OS_NETBSD;
 			break;
 		default:
@@ -1148,29 +1153,6 @@ woptions(struct curparse *curp, char *arg)
 	return 1;
 }
 
-static void
-mmsg(enum mandocerr t, enum mandoclevel lvl,
-		const char *file, int line, int col, const char *msg)
-{
-	const char	*mparse_msg;
-
-	fprintf(mmsg_stream, "%s: %s:", getprogname(),
-	    file == NULL ? "<stdin>" : file);
-
-	if (line)
-		fprintf(mmsg_stream, "%d:%d:", line, col + 1);
-
-	fprintf(mmsg_stream, " %s", mparse_strlevel(lvl));
-
-	if ((mparse_msg = mparse_strerror(t)) != NULL)
-		fprintf(mmsg_stream, ": %s", mparse_msg);
-
-	if (msg)
-		fprintf(mmsg_stream, ": %s", msg);
-
-	fputc('\n', mmsg_stream);
-}
-
 static pid_t
 spawn_pager(struct tag_files *tag_files)
 {
@@ -1179,8 +1161,10 @@ spawn_pager(struct tag_files *tag_files)
 	char		*argv[MAX_PAGER_ARGS];
 	const char	*pager;
 	char		*cp;
+#if HAVE_LESS_T
 	size_t		 cmdlen;
-	int		 argc;
+#endif
+	int		 argc, use_ofn;
 	pid_t		 pager_pid;
 
 	pager = getenv("MANPAGER");
@@ -1196,7 +1180,7 @@ spawn_pager(struct tag_files *tag_files)
 	 */
 
 	argc = 0;
-	while (argc + 4 < MAX_PAGER_ARGS) {
+	while (argc + 5 < MAX_PAGER_ARGS) {
 		argv[argc++] = cp;
 		cp = strchr(cp, ' ');
 		if (cp == NULL)
@@ -1210,14 +1194,23 @@ spawn_pager(struct tag_files *tag_files)
 
 	/* For less(1), use the tag file. */
 
+	use_ofn = 1;
+#if HAVE_LESS_T
 	if ((cmdlen = strlen(argv[0])) >= 4) {
 		cp = argv[0] + cmdlen - 4;
 		if (strcmp(cp, "less") == 0) {
 			argv[argc++] = mandoc_strdup("-T");
 			argv[argc++] = tag_files->tfn;
+			if (tag_files->tagname != NULL) {
+				argv[argc++] = mandoc_strdup("-t");
+				argv[argc++] = tag_files->tagname;
+				use_ofn = 0;
+			}
 		}
 	}
-	argv[argc++] = tag_files->ofn;
+#endif
+	if (use_ofn)
+		argv[argc++] = tag_files->ofn;
 	argv[argc] = NULL;
 
 	switch (pager_pid = fork()) {
