@@ -1,7 +1,7 @@
-/*	$Id: mdoc_term.c,v 1.367 2018/04/11 17:11:13 schwarze Exp $ */
+/*	$Id: mdoc_term.c,v 1.372 2019/01/04 03:39:01 schwarze Exp $ */
 /*
  * Copyright (c) 2008, 2009, 2010, 2011 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2010, 2012-2018 Ingo Schwarze <schwarze@openbsd.org>
+ * Copyright (c) 2010, 2012-2019 Ingo Schwarze <schwarze@openbsd.org>
  * Copyright (c) 2013 Franco Fichtner <franco@lastsummer.de>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -29,7 +29,6 @@
 #include <string.h>
 
 #include "mandoc_aux.h"
-#include "mandoc.h"
 #include "roff.h"
 #include "mdoc.h"
 #include "out.h"
@@ -47,7 +46,7 @@ struct	termpair {
 		  const struct roff_meta *meta, \
 		  struct roff_node *n
 
-struct	termact {
+struct	mdoc_term_act {
 	int	(*pre)(DECL_ARGS);
 	void	(*post)(DECL_ARGS);
 };
@@ -84,6 +83,7 @@ static	void	  termp_xx_post(DECL_ARGS);
 
 static	int	  termp__a_pre(DECL_ARGS);
 static	int	  termp__t_pre(DECL_ARGS);
+static	int	  termp_abort_pre(DECL_ARGS);
 static	int	  termp_an_pre(DECL_ARGS);
 static	int	  termp_ap_pre(DECL_ARGS);
 static	int	  termp_bd_pre(DECL_ARGS);
@@ -124,7 +124,7 @@ static	int	  termp_vt_pre(DECL_ARGS);
 static	int	  termp_xr_pre(DECL_ARGS);
 static	int	  termp_xx_pre(DECL_ARGS);
 
-static	const struct termact __termacts[MDOC_MAX - MDOC_Dd] = {
+static const struct mdoc_term_act mdoc_term_acts[MDOC_MAX - MDOC_Dd] = {
 	{ NULL, NULL }, /* Dd */
 	{ NULL, NULL }, /* Dt */
 	{ NULL, NULL }, /* Os */
@@ -159,7 +159,7 @@ static	const struct termact __termacts[MDOC_MAX - MDOC_Dd] = {
 	{ termp_nd_pre, NULL }, /* Nd */
 	{ termp_nm_pre, termp_nm_post }, /* Nm */
 	{ termp_quote_pre, termp_quote_post }, /* Op */
-	{ termp_ft_pre, NULL }, /* Ot */
+	{ termp_abort_pre, NULL }, /* Ot */
 	{ termp_under_pre, NULL }, /* Pa */
 	{ termp_ex_pre, NULL }, /* Rv */
 	{ NULL, NULL }, /* St */
@@ -232,7 +232,7 @@ static	const struct termact __termacts[MDOC_MAX - MDOC_Dd] = {
 	{ termp_under_pre, NULL }, /* Fr */
 	{ NULL, NULL }, /* Ud */
 	{ NULL, termp_lb_post }, /* Lb */
-	{ termp_pp_pre, NULL }, /* Lp */
+	{ termp_abort_pre, NULL }, /* Lp */
 	{ termp_lk_pre, NULL }, /* Lk */
 	{ termp_under_pre, NULL }, /* Mt */
 	{ termp_quote_pre, termp_quote_post }, /* Brq */
@@ -246,13 +246,12 @@ static	const struct termact __termacts[MDOC_MAX - MDOC_Dd] = {
 	{ NULL, termp____post }, /* %U */
 	{ NULL, NULL }, /* Ta */
 };
-static	const struct termact *const termacts = __termacts - MDOC_Dd;
 
 static	int	 fn_prio;
 
 
 void
-terminal_mdoc(void *arg, const struct roff_man *mdoc)
+terminal_mdoc(void *arg, const struct roff_meta *mdoc)
 {
 	struct roff_node	*n;
 	struct termp		*p;
@@ -270,8 +269,7 @@ terminal_mdoc(void *arg, const struct roff_man *mdoc)
 			if (n->tok == MDOC_Sh && n->sec == SEC_SYNOPSIS) {
 				if (n->child->next->child != NULL)
 					print_mdoc_nodelist(p, NULL,
-					    &mdoc->meta,
-					    n->child->next->child);
+					    mdoc, n->child->next->child);
 				term_newln(p);
 				break;
 			}
@@ -281,8 +279,7 @@ terminal_mdoc(void *arg, const struct roff_man *mdoc)
 		save_defindent = p->defindent;
 		if (p->defindent == 0)
 			p->defindent = 5;
-		term_begin(p, print_mdoc_head, print_mdoc_foot,
-		    &mdoc->meta);
+		term_begin(p, print_mdoc_head, print_mdoc_foot, mdoc);
 		while (n != NULL &&
 		    (n->type == ROFFT_COMMENT ||
 		     n->flags & NODE_NOPRT))
@@ -290,7 +287,7 @@ terminal_mdoc(void *arg, const struct roff_man *mdoc)
 		if (n != NULL) {
 			if (n->tok != MDOC_Sh)
 				term_vspace(p);
-			print_mdoc_nodelist(p, NULL, &mdoc->meta, n);
+			print_mdoc_nodelist(p, NULL, mdoc, n);
 		}
 		term_end(p);
 		p->defindent = save_defindent;
@@ -310,9 +307,23 @@ print_mdoc_nodelist(DECL_ARGS)
 static void
 print_mdoc_node(DECL_ARGS)
 {
-	int		 chld;
+	const struct mdoc_term_act *act;
 	struct termpair	 npair;
 	size_t		 offset, rmargin;
+	int		 chld;
+
+	/*
+	 * In no-fill mode, break the output line at the beginning
+	 * of new input lines except after \c, and nowhere else.
+	 */
+
+	if (n->flags & NODE_NOFILL) {
+		if (n->flags & NODE_LINE &&
+		    (p->flags & TERMP_NONEWLINE) == 0)
+			term_newln(p);
+		p->flags |= TERMP_BRNEVER;
+	} else
+		p->flags &= ~TERMP_BRNEVER;
 
 	if (n->type == ROFFT_COMMENT || n->flags & NODE_NOPRT)
 		return;
@@ -343,9 +354,22 @@ print_mdoc_node(DECL_ARGS)
 
 	switch (n->type) {
 	case ROFFT_TEXT:
-		if (*n->string == ' ' && n->flags & NODE_LINE &&
-		    (p->flags & TERMP_NONEWLINE) == 0)
-			term_newln(p);
+		if (n->flags & NODE_LINE) {
+			switch (*n->string) {
+			case '\0':
+				if (p->flags & TERMP_NONEWLINE)
+					term_newln(p);
+				else
+					term_vspace(p);
+				return;
+			case ' ':
+				if ((p->flags & TERMP_NONEWLINE) == 0)
+					term_newln(p);
+				break;
+			default:
+				break;
+			}
+		}
 		if (NODE_DELIMC & n->flags)
 			p->flags |= TERMP_NOSPACE;
 		term_word(p, n->string);
@@ -370,10 +394,10 @@ print_mdoc_node(DECL_ARGS)
 			return;
 		}
 		assert(n->tok >= MDOC_Dd && n->tok < MDOC_MAX);
-		if (termacts[n->tok].pre != NULL &&
+		act = mdoc_term_acts + (n->tok - MDOC_Dd);
+		if (act->pre != NULL &&
 		    (n->end == ENDBODY_NOT || n->child != NULL))
-			chld = (*termacts[n->tok].pre)
-				(p, &npair, meta, n);
+			chld = (*act->pre)(p, &npair, meta, n);
 		break;
 	}
 
@@ -391,9 +415,9 @@ print_mdoc_node(DECL_ARGS)
 	case ROFFT_EQN:
 		break;
 	default:
-		if (termacts[n->tok].post == NULL || n->flags & NODE_ENDED)
+		if (act->post == NULL || n->flags & NODE_ENDED)
 			break;
-		(void)(*termacts[n->tok].post)(p, &npair, meta, n);
+		(void)(*act->post)(p, &npair, meta, n);
 
 		/*
 		 * Explicit end tokens not only call the post
@@ -1420,8 +1444,6 @@ termp_fa_pre(DECL_ARGS)
 static int
 termp_bd_pre(DECL_ARGS)
 {
-	size_t			 lm, len;
-	struct roff_node	*nn;
 	int			 offset;
 
 	if (n->type == ROFFT_BLOCK) {
@@ -1447,66 +1469,19 @@ termp_bd_pre(DECL_ARGS)
 			p->tcol->offset += offset;
 	}
 
-	/*
-	 * If -ragged or -filled are specified, the block does nothing
-	 * but change the indentation.  If -unfilled or -literal are
-	 * specified, text is printed exactly as entered in the display:
-	 * for macro lines, a newline is appended to the line.  Blank
-	 * lines are allowed.
-	 */
-
-	if (n->norm->Bd.type != DISP_literal &&
-	    n->norm->Bd.type != DISP_unfilled &&
-	    n->norm->Bd.type != DISP_centered)
-		return 1;
-
-	if (n->norm->Bd.type == DISP_literal) {
+	switch (n->norm->Bd.type) {
+	case DISP_literal:
 		term_tab_set(p, NULL);
 		term_tab_set(p, "T");
 		term_tab_set(p, "8n");
+		break;
+	case DISP_centered:
+		p->flags |= TERMP_CENTER;
+		break;
+	default:
+		break;
 	}
-
-	lm = p->tcol->offset;
-	p->flags |= TERMP_BRNEVER;
-	for (nn = n->child; nn != NULL; nn = nn->next) {
-		if (n->norm->Bd.type == DISP_centered) {
-			if (nn->type == ROFFT_TEXT) {
-				len = term_strlen(p, nn->string);
-				p->tcol->offset = len >= p->tcol->rmargin ?
-				    0 : lm + len >= p->tcol->rmargin ?
-				    p->tcol->rmargin - len :
-				    (lm + p->tcol->rmargin - len) / 2;
-			} else
-				p->tcol->offset = lm;
-		}
-		print_mdoc_node(p, pair, meta, nn);
-		/*
-		 * If the printed node flushes its own line, then we
-		 * needn't do it here as well.  This is hacky, but the
-		 * notion of selective eoln whitespace is pretty dumb
-		 * anyway, so don't sweat it.
-		 */
-		if (nn->tok < ROFF_MAX)
-			continue;
-		switch (nn->tok) {
-		case MDOC_Sm:
-		case MDOC_Bl:
-		case MDOC_D1:
-		case MDOC_Dl:
-		case MDOC_Lp:
-		case MDOC_Pp:
-			continue;
-		default:
-			break;
-		}
-		if (p->flags & TERMP_NONEWLINE ||
-		    (nn->next && ! (nn->next->flags & NODE_LINE)))
-			continue;
-		term_flushln(p);
-		p->flags |= TERMP_NOSPACE;
-	}
-	p->flags &= ~TERMP_BRNEVER;
-	return 0;
+	return 1;
 }
 
 static void
@@ -1514,12 +1489,14 @@ termp_bd_post(DECL_ARGS)
 {
 	if (n->type != ROFFT_BODY)
 		return;
-	if (DISP_literal == n->norm->Bd.type ||
-	    DISP_unfilled == n->norm->Bd.type)
+	if (n->norm->Bd.type == DISP_unfilled ||
+	    n->norm->Bd.type == DISP_literal)
 		p->flags |= TERMP_BRNEVER;
 	p->flags |= TERMP_NOSPACE;
 	term_newln(p);
 	p->flags &= ~TERMP_BRNEVER;
+	if (n->norm->Bd.type == DISP_centered)
+		p->flags &= ~TERMP_CENTER;
 }
 
 static int
@@ -2098,3 +2075,9 @@ termp_tag_pre(DECL_ARGS)
 		tag_put(n->child->string, 1, p->line);
 	return 1;
 }
+
+static int
+termp_abort_pre(DECL_ARGS)
+{
+	abort();
+}
